#!/usr/bin/env python3

""" Example of announcing a service (in this case, a fake HTTP server) """

import argparse
import logging
import socket
import time
from time import sleep

from zeroconf import IPVersion, ServiceInfo, Zeroconf


def get_host_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('8.8.8.8', 80))  # 114.114.114.114也是dns地址
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip


def mdns_register():
    ip_version = IPVersion.V4Only
    desc = {'path': '/~paulsm/'}

    info = ServiceInfo(
        "_http._tcp.local.",
        "EasySmartSever._http._tcp.local.",
        addresses=[socket.inet_aton(get_host_ip())],
        port=80,
        properties=desc,
        server="mqttserver.local.",
    )

    zeroconf = Zeroconf(ip_version=ip_version)
    print("Registration of a service")
    zeroconf.register_service(info)
    # try:
    #     while True:
    #         sleep(0.1)
    # except KeyboardInterrupt:
    #     pass
    # finally:
    #     print("Unregistering...")
    #     zeroconf.unregister_service(info)
    #     zeroconf.close()
    time.sleep(1)



if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    mdns_register()



import json
from paho.mqtt.client import MQTT_LOG_INFO, MQTT_LOG_NOTICE, MQTT_LOG_WARNING, MQTT_LOG_ERR, MQTT_LOG_DEBUG

def on_log(client, userdata, level, buf):
    if level == MQTT_LOG_INFO:
        head = 'INFO'
    elif level == MQTT_LOG_NOTICE:
        head = 'NOTICE'
    elif level == MQTT_LOG_WARNING:
        head = 'WARN'
    elif level == MQTT_LOG_ERR:
        head = 'ERR'
    elif level == MQTT_LOG_DEBUG:
        head = 'DEBUG'
    else:
        head = level
    print('%s: %s' % (head, buf))


def on_connect(client, userdata, flags, rc):
    print('Connected with result code ' + str(rc))
    # client.subscribe(topic, 0)


def on_message(client, userdata, msg):
    print('topic:' + msg.topic + ' ' + str(msg.payload))
    try:
        data = json.loads(msg.payload)
    except:
        data = str(msg.payload)
    print(f'{data}')


def on_disconnect(client, userdata, rc):
    if rc != 0:
        print('Unexpected disconnection %s' % rc)



import asyncio
import json
import logging
from asyncio import events
from typing import Optional

from aiohttp import web
from aiohttp.log import access_logger
from aiohttp.web import _run_app
from aiohttp.web_log import AccessLogger

from easysmart import Manager


async def run_app(
        app,
        *,
        host=None,
        port=None,
        path=None,
        sock=None,
        shutdown_timeout: float = 60.0,
        keepalive_timeout: float = 75.0,
        ssl_context=None,
        print=print,
        backlog: int = 128,
        access_log_class=AccessLogger,
        access_log_format: str = AccessLogger.LOG_FORMAT,
        access_log=access_logger,
        handle_signals: bool = True,
        reuse_address: Optional[bool] = None,
        reuse_port: Optional[bool] = None,
        handler_cancellation: bool = False,
        loop: Optional[asyncio.AbstractEventLoop] = None,
):
    """Run an app locally"""
    if loop is None:
        loop = asyncio.new_event_loop()

    # Configure if and only if in debugging mode and using the default logger
    if loop.get_debug() and access_log and access_log.name == "aiohttp.access":
        if access_log.level == logging.NOTSET:
            access_log.setLevel(logging.DEBUG)
        if not access_log.hasHandlers():
            access_log.addHandler(logging.StreamHandler())

    main_task = loop.create_task(
        _run_app(
            app,
            host=host,
            port=port,
            path=path,
            sock=sock,
            shutdown_timeout=shutdown_timeout,
            keepalive_timeout=keepalive_timeout,
            ssl_context=ssl_context,
            print=print,
            backlog=backlog,
            access_log_class=access_log_class,
            access_log_format=access_log_format,
            access_log=access_log,
            handle_signals=handle_signals,
            reuse_address=reuse_address,
            reuse_port=reuse_port,
            handler_cancellation=handler_cancellation,
        )
    )

    await main_task


def build_response(msg=None, data=None, code=200, success=True, ):
    res = {
        'code': code,
        'success': success,
    }
    if data is not None:
        res['data'] = data
    if msg is not None:
        res['msg'] = msg
    return web.json_response(res)


class WebServer:

    def __init__(self, manager: 'Manager'):
        self.manager = manager

    async def web_start(self):
        print('启动http服务器')
        app = web.Application()
        app.add_routes([web.get('/', self.handle),
                        web.get('/devices/', self.devices_view),
                        web.get('/device/{mac}/', self.device_get_view),
                        web.post('/device/{mac}/', self.device_set_view),
                        web.post('/device/set/{filter}/{detail}/', self.device_group_set_view),
                        web.post('/device/act/{filter}/{detail}/', self.device_group_act_view),])
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, 'localhost', 8555)
        await site.start()
        # await run_app(app, loop=asyncio.get_event_loop())

    async def handler(self, request):
        return build_response(msg='运行中')

    async def handle(self, request):
        name = request.match_info.get('name', "Anonymous")
        return build_response(msg=f'hello {name}')

    async def devices_view(self, request):
        devices = self.manager.devices
        d = []
        for mac, device in devices.items():
            d.append(device.to_dict())
        return build_response(data=d)

    async def device_get_view(self, request):
        mac = request.match_info.get('mac', "Anonymous")
        if mac in self.manager.devices:
            device = self.manager.devices[mac]
        else:
            return web.Response(text=json.dumps({'error': 'device not found'}))
        return build_response(data=device.to_dict())

    async def device_set_view(self, request):
        mac = request.match_info.get('mac', "Anonymous")
        if mac in self.manager.devices:
            device = self.manager.devices[mac]
        else:
            return build_response(success=False, msg='device not found')
        data = await request.post()
        for k, v in data.items():
            asyncio.gather(device.set_property(k, v))
        return build_response()

    async def device_group_set_view(self, request):
        filter = request.match_info.get('filter', "")
        detail = request.match_info.get('detail', "")

        devices = await self.device_filter(detail, filter)
        if filter == "":
            return build_response(success=False, msg='filter is empty')
        data = await request.post()
        for device in devices:
            for k, v in data.items():
                asyncio.gather(device.set_property(k, v))

        res = {
            'affected_devices': [device.mac for device in devices],
        }
        return build_response(data=res)

    async def device_filter(self, detail, filter):
        if filter == 'all':
            devices = self.manager.devices
        elif filter == '':
            return []
        else:
            devices = []
            for mac, device in self.manager.devices.items():
                device_detail = getattr(device, filter)
                if device_detail == detail:
                    devices.append(device)
        return devices

    async def device_group_act_view(self, request):
        filter = request.match_info.get('filter', "")
        detail = request.match_info.get('detail', "")

        devices = await self.device_filter(detail, filter)
        data = await request.post()
        new_data = {}
        for k, v in data.items():
            value = v
            if str(value).isdigit():
                value = int(value)
            elif '.' in str(value) and str(value).replace('.', '').isdigit():
                value = float(value)
            new_data[k] = value

        for device in devices:
            asyncio.create_task(device.publish(new_data))

        res = {
            'affected_devices': [device.mac for device in devices],
        }
        return build_response(data=res)


if __name__ == '__main__':
    web_manager = WebServer(None)
    loop = asyncio.get_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(web_manager.web_start())
    print('web server start')
    loop.run_forever()


import time
import asyncio

async def say(text, sec):
    await asyncio.sleep(sec)
    print(text)

async def main():
    print('start main')
    start = time.monotonic()
    await asyncio.gather(
        say("First", 2),
        say("Second", 1),
    )
    end = time.monotonic()
    print(f"Elapsed: {end-start}")


from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy

set_event_loop_policy(WindowsSelectorEventLoopPolicy())
try:
    loop = asyncio.get_event_loop()
except:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
main_co = main()
print(main_co)
asyncio.run(main_co)


import asyncio
import json
import os
import threading
import time
import traceback
import warnings

import aiomqtt

from paho.mqtt.client import MQTT_LOG_INFO, MQTT_LOG_NOTICE, MQTT_LOG_WARNING, MQTT_LOG_ERR, MQTT_LOG_DEBUG
from paho.mqtt import client as mqtt
from easysmart.device.base_device import BaseDevice


def on_log(client, userdata, level, buf):
    if level == MQTT_LOG_INFO:
        head = 'INFO'
    elif level == MQTT_LOG_NOTICE:
        head = 'NOTICE'
    elif level == MQTT_LOG_WARNING:
        head = 'WARN'
    elif level == MQTT_LOG_ERR:
        head = 'ERR'
    elif level == MQTT_LOG_DEBUG:
        head = 'DEBUG'
    else:
        head = level
    print('%s: %s' % (head, buf))


def on_connect(client, userdata, flags, rc):
    print('Connected with result code ' + str(rc))
    # client.subscribe(topic, 0)


def on_message(client, userdata, msg):
    print('topic:' + msg.topic + ' ' + str(msg.payload))
    try:
        data = json.loads(msg.payload)
    except:
        data = str(msg.payload)
    print(f'{data}')


def on_disconnect(client, userdata, rc):
    if rc != 0:
        print('Unexpected disconnection %s' % rc)


class Manager:
    client = None
    automation = None

    def __init__(self, on_message_cb=None, on_device_disconnect_cb=None):
        self.client_id = 'MQTT_MAIN' + os.urandom(6).hex()

        self.on_message_cb = on_message_cb
        self.on_device_disconnect_cb = on_device_disconnect_cb

        self.devices = {}

    def subscribe(self, *args, **kwargs):
        return self.client.subscribe(*args, **kwargs)

    async def publish(self, topic, payload=None, qos=0, retain=False, properties=None):
        print(f'publish: {topic} {payload}')
        await self.client.publish(topic=topic, payload=payload, qos=qos, retain=retain, properties=properties)
        # return self.client.publish(topic, payload, qos, retain, properties)

    def loop_forever(self):
        self.client.loop_start()
        self.thread_main()

    def loop_start(self):
        # self._thread_terminate = False
        # self._thread = threading.Thread(target=self.thread_main)
        # self._thread.daemon = True
        # self._thread.start()
        return self.client.loop_start()

    async def thread_main(self):
        """
        主线程
        :return: None
        """
        while True:
            await asyncio.sleep(1)
            await self._seconds_work()

    async def _seconds_work(self):
        """
        每秒被调用一次
        :return: None
        """
        # print('seconds work')
        # print 当前设备列表
        # print('devices:')
        need_del = []
        for k, v in self.devices.items():
            # print(f'{k}: {v}')
            if time.time() - v.last_active_time > 30:
                print(f'device {k} offline')
                need_del.append(k)
        for k in need_del:
            self.devices.pop(k)

    async def async_loop_start(self):
        print('启动主服务循环')
        fail_num = 0
        debug = True
        if debug: warnings.warn('debug mode is on')
        loop = asyncio.get_event_loop()
        loop.create_task(self.thread_main())
        while True:
            try:
                await self._async_loop_start()
            except aiomqtt.error.MqttError as e:
                print(f'[{fail_num}]async_loop_start connect error: {e}')
                fail_num += 1
                await asyncio.sleep(3)
            except Exception as e:
                print(f'[{fail_num}]async_loop_start error: {e}')
                if debug: traceback.print_exc()
                fail_num += 1
                await asyncio.sleep(3)

    async def _async_loop_start(self):
        async with aiomqtt.Client("easysmart.local") as self.client:
            while True:
                try:
                    print('sub #')
                    await self.client.subscribe("/#")
                    print('sub /dpub/#')
                    await self.client.subscribe("/dpub/#")
                    print('sub /all/')
                    await self.client.subscribe("/all/")
                    break
                except Exception as e:
                    print(f'async_loop_start error: {e}')
                    await asyncio.sleep(1)
                    raise e
            print('sub success')
            async with self.client.messages() as messages:
                async for message in messages:
                    payload = message.payload.decode('utf-8')
                    topic = message.topic
                    print(f'async loop: {topic} {payload}')
                    try:
                        data = json.loads(payload)
                    except:
                        data = str(payload)
                        print(f'json.loads error: {data}')
                        continue
                    print(f'{data}')
                    # await self.msg_process(message, data)
                    asyncio.create_task(self.msg_process(message, data))

    def load_auto_config(self):
        if self.automation:
            ...

    def init_sync_mqtt_client(self):
        self.client = mqtt.Client(self.client_id, protocol=mqtt.MQTTv311, clean_session=False)
        self.client.on_log = on_log
        self.client.on_connect = on_connect
        self.client.on_message = self.on_message
        self.client.on_disconnect = on_disconnect
        self.client.connect('easysmart.local', 1883, 60)
        self.client.loop_start()
        self.client.subscribe('#', 0)
        self.client.subscribe('/dpub/#', 1)
        self.client.subscribe('/all/', 1)
        self.client.subscribe('/test/', 1)
        self.client.publish('/test', 'hello world', 0)

    async def msg_process(self, msg, data):
        await self._msg_process(msg, data)
        if self.on_message_cb is not None:
            asyncio.gather(self.on_message_cb(msg, data))

    async def _msg_process(self, msg, data):
        topic = msg.topic.value
        # 如果topic形式为 /dpub/{mac}
        if topic.startswith('/dpub/'):
            mac = topic[6:]
            if mac not in self.devices:
                if data.get('method') == 'report':
                    self.devices[mac] = BaseDevice(mac, data.get('device_type', ''), mqtt_client=self.client)
                    print(f'new device {mac}')
                else:
                    print(f'unknown device {mac} 等待该设备的report信息')
                    return
            method = data.get('method')
            if method == 'report':
                new_data = data.copy()
                new_data.pop('method')
                for k, v in new_data.items():
                    self.devices[mac].update(k, v)
            elif method == 'update':
                key = data.get('key')
                value = data.get('value')
                if key and value:
                    self.devices[mac].update(key, value)

    def get_device(self, mac=None, device_type=None):
        devices = []
        for k, v in self.devices.items():
            if mac and mac != k:
                continue
            if device_type and device_type != v.device_type:
                continue
            devices.append(v)
        return devices


if __name__ == '__main__':
    m = Manager()
    m.loop_start()



import os, sys, pathlib
import time
from pathlib import Path
import asyncio
import aiohttp
import zipfile
from tqdm.asyncio import tqdm



async def download_file(url, save_path):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            total_size = int(response.headers.get('Content-Length', 0))
            with open(save_path, 'wb') as f:
                with tqdm(total=total_size, desc=str(save_path), unit='B', unit_scale=True) as pbar:
                    while True:
                        chunk = await response.content.read(1024)
                        if not chunk:
                            break
                        f.write(chunk)
                        pbar.update(len(chunk))
    return save_path

async def download_emqx_server(root_path):
    print(f'system is {sys.platform}')
    if sys.platform == 'win32':
        pass
    else:
        raise NotImplementedError('only support windows now')

    download_url = r'https://packages.emqx.io/emqx-ce/v5.3.1/emqx-5.3.1-windows-amd64.zip'
    # download emqx server to root_path/emqx.zip
    emqx_zip_path = root_path.joinpath(Path(r'emqx.zip'))
    print(f'emqx zip path is {emqx_zip_path}')
    await download_file(download_url, emqx_zip_path)
    print('download emqx server success')
    # unzip emqx.zip to root_path/emqx
    emqx_path = root_path.joinpath(Path(r'emqx'))
    print(f'emqx path is {emqx_path}')
    with zipfile.ZipFile(emqx_zip_path, 'r') as zip_ref:
        zip_ref.extractall(emqx_path)


async def start_emqx_server(root_path):
    print(f'system is {sys.platform}')
    if sys.platform == 'win32':
        pass
    else:
        raise NotImplementedError('only support windows now')

    # download emqx server if not exists
    emqx_path = root_path.joinpath(Path(r'emqx\bin\emqx_ctl'))
    if os.path.exists(emqx_path):
        print(f'emqx path is {emqx_path}')
    else:
        await download_emqx_server(root_path)

    while not await check_emqx_server(root_path):
        print(f'starting emqx server')
        # if path not exists, create it
        emqx_path = root_path.joinpath(Path(r'emqx\bin\emqx'))
        if os.path.exists(emqx_path):
            print(f'emqx path is {emqx_path}')
        else:
            raise FileNotFoundError(f'emqx path {emqx_path} not exists')
        # start emqx by run "emqx\bin\emqx start"
        os.system(f"{emqx_path} start")
        print('emqx server started')
        await asyncio.sleep(3)


async def check_emqx_server(root_path):
    print(f'check emqx server')
    # if path not exists, create it
    emqx_path = root_path.joinpath(Path(r'emqx\bin\emqx_ctl'))
    if os.path.exists(emqx_path):
        print(f'emqx path is {emqx_path}')
    else:
        raise FileNotFoundError(f'emqx path {emqx_path} not exists')
    # 检测emqx是否正在运行
    # run "emqx\bin\emqx status" and get the result
    result = os.popen(f"{emqx_path} status").read()
    print(f'emqx status is {result}')
    if 'is starting' in result:
        await asyncio.sleep(3)
        return await check_emqx_server(root_path)
    if 'is started' in result:
        print('emqx server is running')
        return True
    else:
        print('emqx server is not running')
        return False

def sync_check_emqx_server(root_path):
    print(f'check emqx server')
    # if path not exists, create it
    emqx_path = root_path.joinpath(Path(r'emqx\bin\emqx_ctl'))
    if os.path.exists(emqx_path):
        print(f'emqx path is {emqx_path}')
    else:
        raise FileNotFoundError(f'emqx path {emqx_path} not exists')
    # 检测emqx是否正在运行
    # run "emqx\bin\emqx status" and get the result
    result = os.popen(f"{emqx_path} status").read()
    print(f'emqx status is {result}')
    if 'is starting' in result:
        return 'starting'
    if 'is started' in result:
        print('emqx server is running')
        return 'running'
    else:
        print('emqx server is not running')
        return False
